@@ -26,10 +26,12 @@
 
 #include "config.h"
 
+#include <stdbool.h>
 #include "linux/input.h"
 #include <libevdev/libevdev.h>
 
 #include "libinput-private.h"
+#include "timer.h"
 
 enum evdev_event_type {
 	EVDEV_NONE,
@@ -48,6 +50,12 @@ enum evdev_device_seat_capability {
 	EVDEV_DEVICE_TOUCH = (1 << 2)
 };
 
+enum evdev_device_tags {
+	EVDEV_TAG_EXTERNAL_MOUSE = (1 << 0),
+	EVDEV_TAG_INTERNAL_TOUCHPAD = (1 << 1),
+	EVDEV_TAG_TRACKPOINT = (1 << 2),
+};
+
 struct mt_slot {
 	int32_t seat_slot;
 	int32_t x, y;
@@ -60,10 +68,10 @@ struct evdev_device {
 
 	struct evdev_dispatch *dispatch;
 	struct libevdev *evdev;
+	struct udev_device *udev_device;
 	char *output_name;
-	char *devnode;
-	char *sysname;
 	const char *devname;
+	bool was_removed;
 	int fd;
 	struct {
 		const struct input_absinfo *absinfo_x, *absinfo_y;
@@ -89,21 +97,62 @@ struct evdev_device {
 		int dx, dy;
 	} rel;
 
+	struct {
+		struct libinput_timer timer;
+		struct libinput_device_config_scroll_method config;
+		/* Currently enabled method, button */
+		enum libinput_config_scroll_method method;
+		uint32_t button;
+		/* set during device init, used at runtime to delay changes
+		 * until all buttons are up */
+		enum libinput_config_scroll_method want_method;
+		uint32_t want_button;
+		/* Checks if buttons are down and commits the setting */
+		void (*change_scroll_method)(struct evdev_device *device);
+		bool button_scroll_active;
+		double threshold;
+		uint32_t direction;
+		double buildup_vertical;
+		double buildup_horizontal;
+
+		struct libinput_device_config_natural_scroll config_natural;
+		/* set during device init if we want natural scrolling,
+		 * used at runtime to enable/disable the feature */
+		bool natural_scrolling_enabled;
+	} scroll;
+
 	enum evdev_event_type pending_event;
 	enum evdev_device_seat_capability seat_caps;
+	enum evdev_device_tags tags;
 
 	int is_mt;
+	int suspended;
 
 	struct {
+		struct libinput_device_config_accel config;
 		struct motion_filter *filter;
 	} pointer;
 
 	/* Bitmask of pressed keys used to ignore initial release events from
 	 * the kernel. */
-	unsigned long key_mask[NLONGS(KEY_CNT)];
+	unsigned long hw_key_mask[NLONGS(KEY_CNT)];
 	/* Key counter used for multiplexing button events internally in
 	 * libinput. */
 	uint8_t key_count[KEY_CNT];
+
+	struct {
+		struct libinput_device_config_left_handed config_left_handed;
+		/* left-handed currently enabled */
+		bool left_handed;
+		/* set during device init if we want left_handed config,
+		 * used at runtime to delay the effect until buttons are up */
+		bool want_left_handed;
+		/* Checks if buttons are down and commits the setting */
+		void (*change_to_left_handed)(struct evdev_device *device);
+	} buttons;
+
+	int dpi; /* HW resolution */
+	struct ratelimit syn_drop_limit; /* ratelimit for SYN_DROPPED logging */
 };
 
 #define EVDEV_UNHANDLED_DEVICE ((struct evdev_device *) 1)
@@ -119,17 +168,44 @@ struct evdev_dispatch_interface {
 
 	/* Destroy an event dispatch handler and free all its resources. */
 	void (*destroy)(struct evdev_dispatch *dispatch);
+
+	/* A new device was added */
+	void (*device_added)(struct evdev_device *device,
+			     struct evdev_device *added_device);
+
+	/* A device was removed */
+	void (*device_removed)(struct evdev_device *device,
+			       struct evdev_device *removed_device);
+
+	/* A device was suspended */
+	void (*device_suspended)(struct evdev_device *device,
+				 struct evdev_device *suspended_device);
+
+	/* A device was resumed */
+	void (*device_resumed)(struct evdev_device *device,
+			       struct evdev_device *resumed_device);
+
+	/* Tag device with one of EVDEV_TAG */
+	void (*tag_device)(struct evdev_device *device,
+			   struct udev_device *udev_device);
 };
 
 struct evdev_dispatch {
 	struct evdev_dispatch_interface *interface;
 	struct libinput_device_config_calibration calibration;
+
+	struct {
+		struct libinput_device_config_send_events config;
+		enum libinput_config_send_events_mode current_mode;
+	} sendevents;
 };
 
 struct evdev_device *
 evdev_device_create(struct libinput_seat *seat,
-		    const char *devnode,
-		    const char *sysname);
+		    struct udev_device *device);
+
+int
+evdev_device_init_pointer_acceleration(struct evdev_device *device);
 
 struct evdev_dispatch *
 evdev_touchpad_create(struct evdev_device *device);
@@ -137,9 +213,6 @@ evdev_touchpad_create(struct evdev_device *device);
 struct evdev_dispatch *
 evdev_mt_touchpad_create(struct evdev_device *device);
 
-void
-evdev_device_proces_event(struct libinput_event *event);
-
 void
 evdev_device_led_update(struct evdev_device *device, enum libinput_led leds);
 
@@ -161,6 +234,9 @@ evdev_device_get_id_product(struct evdev_device *device);
 unsigned int
 evdev_device_get_id_vendor(struct evdev_device *device);
 
+struct udev_device *
+evdev_device_get_udev_device(struct evdev_device *device);
+
 void
 evdev_device_set_default_calibration(struct evdev_device *device,
 				     const float calibration[6]);
@@ -177,6 +253,9 @@ evdev_device_get_size(struct evdev_device *device,
 		      double *w,
 		      double *h);
 
+int
+evdev_device_has_button(struct evdev_device *device, uint32_t code);
+
 double
 evdev_device_transform_x(struct evdev_device *device,
 			 double x,
@@ -186,6 +265,17 @@ double
 evdev_device_transform_y(struct evdev_device *device,
 			 double y,
 			 uint32_t height);
+int
+evdev_device_suspend(struct evdev_device *device);
+
+int
+evdev_device_resume(struct evdev_device *device);
+
+void
+evdev_notify_suspended_device(struct evdev_device *device);
+
+void
+evdev_notify_resumed_device(struct evdev_device *device);
 
 void
 evdev_keyboard_notify_key(struct evdev_device *device,
@@ -199,6 +289,19 @@ evdev_pointer_notify_button(struct evdev_device *device,
 			    int button,
 			    enum libinput_button_state state);
 
+void
+evdev_init_natural_scroll(struct evdev_device *device);
+
+void
+evdev_post_scroll(struct evdev_device *device,
+		  uint64_t time,
+		  double dx,
+		  double dy);
+
+
+void
+evdev_stop_scroll(struct evdev_device *device, uint64_t time);
+
 void
 evdev_device_remove(struct evdev_device *device);
 
@@ -212,4 +315,21 @@ evdev_convert_to_mm(const struct input_absinfo *absinfo, double v)
 	return value/absinfo->resolution;
 }
 
+int
+evdev_init_left_handed(struct evdev_device *device,
+		       void (*change_to_left_handed)(struct evdev_device *));
+
+static inline uint32_t
+evdev_to_left_handed(struct evdev_device *device,
+		     uint32_t button)
+{
+	if (device->buttons.left_handed) {
+		if (button == BTN_LEFT)
+			return BTN_RIGHT;
+		else if (button == BTN_RIGHT)
+			return BTN_LEFT;
+	}
+	return button;
+}
+
 #endif /* EVDEV_H */
