@@ -23,6 +23,8 @@
 #ifndef LIBINPUT_PRIVATE_H
 #define LIBINPUT_PRIVATE_H
 
+#include <errno.h>
+
 #include "linux/input.h"
 
 #include "libinput.h"
@@ -34,6 +36,8 @@ struct libinput_interface_backend {
 	int (*resume)(struct libinput *libinput);
 	void (*suspend)(struct libinput *libinput);
 	void (*destroy)(struct libinput *libinput);
+	int (*device_change_seat)(struct libinput_device *device,
+				  const char *seat_name);
 };
 
 struct libinput {
@@ -99,20 +103,79 @@ struct libinput_device_config_calibration {
 							  float matrix[6]);
 };
 
+struct libinput_device_config_send_events {
+	uint32_t (*get_modes)(struct libinput_device *device);
+	enum libinput_config_status (*set_mode)(struct libinput_device *device,
+						   enum libinput_config_send_events_mode mode);
+	enum libinput_config_send_events_mode (*get_mode)(struct libinput_device *device);
+	enum libinput_config_send_events_mode (*get_default_mode)(struct libinput_device *device);
+};
+
+struct libinput_device_config_accel {
+	int (*available)(struct libinput_device *device);
+	enum libinput_config_status (*set_speed)(struct libinput_device *device,
+						 double speed);
+	double (*get_speed)(struct libinput_device *device);
+	double (*get_default_speed)(struct libinput_device *device);
+};
+
+struct libinput_device_config_natural_scroll {
+	int (*has)(struct libinput_device *device);
+	enum libinput_config_status (*set_enabled)(struct libinput_device *device,
+						   int enabled);
+	int (*get_enabled)(struct libinput_device *device);
+	int (*get_default_enabled)(struct libinput_device *device);
+};
+
+struct libinput_device_config_left_handed {
+	int (*has)(struct libinput_device *device);
+	enum libinput_config_status (*set)(struct libinput_device *device, int left_handed);
+	int (*get)(struct libinput_device *device);
+	int (*get_default)(struct libinput_device *device);
+};
+
+struct libinput_device_config_scroll_method {
+	uint32_t (*get_methods)(struct libinput_device *device);
+	enum libinput_config_status (*set_method)(struct libinput_device *device,
+						  enum libinput_config_scroll_method method);
+	enum libinput_config_scroll_method (*get_method)(struct libinput_device *device);
+	enum libinput_config_scroll_method (*get_default_method)(struct libinput_device *device);
+	enum libinput_config_status (*set_button)(struct libinput_device *device,
+						  uint32_t button);
+	uint32_t (*get_button)(struct libinput_device *device);
+	uint32_t (*get_default_button)(struct libinput_device *device);
+};
+
 struct libinput_device_config {
 	struct libinput_device_config_tap *tap;
 	struct libinput_device_config_calibration *calibration;
+	struct libinput_device_config_send_events *sendevents;
+	struct libinput_device_config_accel *accel;
+	struct libinput_device_config_natural_scroll *natural_scroll;
+	struct libinput_device_config_left_handed *left_handed;
+	struct libinput_device_config_scroll_method *scroll_method;
 };
 
 struct libinput_device {
 	struct libinput_seat *seat;
 	struct list link;
+	struct list event_listeners;
 	void *user_data;
-	int terminated;
 	int refcount;
 	struct libinput_device_config config;
 };
 
+struct libinput_event {
+	enum libinput_event_type type;
+	struct libinput_device *device;
+};
+
+struct libinput_event_listener {
+	struct list link;
+	void (*notify_func)(uint64_t time, struct libinput_event *ev, void *notify_func_data);
+	void *notify_func_data;
+};
+
 typedef void (*libinput_source_dispatch_t)(void *data);
 
 
@@ -120,8 +183,8 @@ typedef void (*libinput_source_dispatch_t)(void *data);
 #define log_info(li_, ...) log_msg((li_), LIBINPUT_LOG_PRIORITY_INFO, __VA_ARGS__)
 #define log_error(li_, ...) log_msg((li_), LIBINPUT_LOG_PRIORITY_ERROR, __VA_ARGS__)
 #define log_bug_kernel(li_, ...) log_msg((li_), LIBINPUT_LOG_PRIORITY_ERROR, "kernel bug: " __VA_ARGS__)
-#define log_bug_libinput(li_, ...) log_msg((li_), LIBINPUT_LOG_PRIORITY_ERROR, "libinput bug: " __VA_ARGS__);
-#define log_bug_client(li_, ...) log_msg((li_), LIBINPUT_LOG_PRIORITY_ERROR, "client bug: " __VA_ARGS__);
+#define log_bug_libinput(li_, ...) log_msg((li_), LIBINPUT_LOG_PRIORITY_ERROR, "libinput bug: " __VA_ARGS__)
+#define log_bug_client(li_, ...) log_msg((li_), LIBINPUT_LOG_PRIORITY_ERROR, "client bug: " __VA_ARGS__)
 
 void
 log_msg(struct libinput *libinput,
@@ -168,6 +231,18 @@ void
 libinput_device_init(struct libinput_device *device,
 		     struct libinput_seat *seat);
 
+void
+libinput_device_add_event_listener(struct libinput_device *device,
+				   struct libinput_event_listener *listener,
+				   void (*notify_func)(
+						uint64_t time,
+						struct libinput_event *event,
+						void *notify_func_data),
+				   void *notify_func_data);
+
+void
+libinput_device_remove_event_listener(struct libinput_event_listener *listener);
+
 void
 notify_added_device(struct libinput_device *device);
 
@@ -176,37 +251,39 @@ notify_removed_device(struct libinput_device *device);
 
 void
 keyboard_notify_key(struct libinput_device *device,
-		    uint32_t time,
+		    uint64_t time,
 		    uint32_t key,
 		    enum libinput_key_state state);
 
 void
 pointer_notify_motion(struct libinput_device *device,
-		      uint32_t time,
+		      uint64_t time,
 		      double dx,
-		      double dy);
+		      double dy,
+		      double dx_noaccel,
+		      double dy_noaccel);
 
 void
 pointer_notify_motion_absolute(struct libinput_device *device,
-			       uint32_t time,
+			       uint64_t time,
 			       double x,
 			       double y);
 
 void
 pointer_notify_button(struct libinput_device *device,
-		      uint32_t time,
+		      uint64_t time,
 		      int32_t button,
 		      enum libinput_button_state state);
 
 void
 pointer_notify_axis(struct libinput_device *device,
-		    uint32_t time,
+		    uint64_t time,
 		    enum libinput_pointer_axis axis,
 		    double value);
 
 void
 touch_notify_touch_down(struct libinput_device *device,
-			uint32_t time,
+			uint64_t time,
 			int32_t slot,
 			int32_t seat_slot,
 			double x,
@@ -214,7 +291,7 @@ touch_notify_touch_down(struct libinput_device *device,
 
 void
 touch_notify_touch_motion(struct libinput_device *device,
-			  uint32_t time,
+			  uint64_t time,
 			  int32_t slot,
 			  int32_t seat_slot,
 			  double x,
@@ -222,11 +299,24 @@ touch_notify_touch_motion(struct libinput_device *device,
 
 void
 touch_notify_touch_up(struct libinput_device *device,
-		      uint32_t time,
+		      uint64_t time,
 		      int32_t slot,
 		      int32_t seat_slot);
 
 void
 touch_notify_frame(struct libinput_device *device,
-		   uint32_t time);
+		   uint64_t time);
+
+static inline uint64_t
+libinput_now(struct libinput *libinput)
+{
+	struct timespec ts = { 0, 0 };
+
+	if (clock_gettime(CLOCK_MONOTONIC, &ts) != 0) {
+		log_error(libinput, "clock_gettime failed: %s\n", strerror(errno));
+		return 0;
+	}
+
+	return ts.tv_sec * 1000ULL + ts.tv_nsec / 1000000;
+}
 #endif /* LIBINPUT_PRIVATE_H */
