@@ -87,6 +87,9 @@ extern struct litest_test_device litest_mouse_device;
 extern struct litest_test_device litest_wacom_touch_device;
 extern struct litest_test_device litest_alps_device;
 extern struct litest_test_device litest_generic_singletouch_device;
+extern struct litest_test_device litest_qemu_tablet_device;
+extern struct litest_test_device litest_xen_virtual_pointer_device;
+extern struct litest_test_device litest_vmware_virtmouse_device;
 
 struct litest_test_device* devices[] = {
 	&litest_synaptics_clickpad_device,
@@ -99,6 +102,9 @@ struct litest_test_device* devices[] = {
 	&litest_wacom_touch_device,
 	&litest_alps_device,
 	&litest_generic_singletouch_device,
+	&litest_qemu_tablet_device,
+	&litest_xen_virtual_pointer_device,
+	&litest_vmware_virtmouse_device,
 	NULL,
 };
 
@@ -159,6 +165,9 @@ litest_add_tcase(struct suite *suite, void *func,
 {
 	struct litest_test_device **dev = devices;
 
+	assert(required >= LITEST_DISABLE_DEVICE);
+	assert(excluded >= LITEST_DISABLE_DEVICE);
+
 	if (required == LITEST_DISABLE_DEVICE &&
 	    excluded == LITEST_DISABLE_DEVICE) {
 		litest_add_tcase_no_device(suite, func);
@@ -223,6 +232,8 @@ litest_add_for_device(const char *name,
 	struct suite *s;
 	struct litest_test_device **dev = devices;
 
+	assert(type < LITEST_NO_DEVICE);
+
 	s = get_suite(name);
 	while (*dev) {
 		if ((*dev)->type == type) {
@@ -558,6 +569,18 @@ litest_add_device_with_overrides(struct libinput *libinput,
 	return d;
 }
 
+struct litest_device *
+litest_add_device(struct libinput *libinput,
+		  enum litest_device_type which)
+{
+	return litest_add_device_with_overrides(libinput,
+						which,
+						NULL,
+						NULL,
+						NULL,
+						NULL);
+}
+
 struct litest_device *
 litest_create_device_with_overrides(enum litest_device_type which,
 				    const char *name_override,
@@ -617,7 +640,12 @@ void
 litest_event(struct litest_device *d, unsigned int type,
 	     unsigned int code, int value)
 {
-	int ret = libevdev_uinput_write_event(d->uinput, type, code, value);
+	int ret;
+
+	if (d->skip_ev_syn && type == EV_SYN && code == SYN_REPORT)
+		return;
+
+	ret = libevdev_uinput_write_event(d->uinput, type, code, value);
 	ck_assert_int_eq(ret, 0);
 }
 
@@ -740,12 +768,18 @@ litest_touch_move_to(struct litest_device *d,
 		     unsigned int slot,
 		     double x_from, double y_from,
 		     double x_to, double y_to,
-		     int steps)
+		     int steps, int sleep_ms)
 {
-	for (int i = 0; i < steps - 1; i++)
+	for (int i = 0; i < steps - 1; i++) {
 		litest_touch_move(d, slot,
 				  x_from + (x_to - x_from)/steps * i,
 				  y_from + (y_to - y_from)/steps * i);
+		if (sleep_ms) {
+			libinput_dispatch(d->libinput);
+			msleep(sleep_ms);
+			libinput_dispatch(d->libinput);
+		}
+	}
 	litest_touch_move(d, slot, x_to, y_to);
 }
 
@@ -763,6 +797,28 @@ litest_button_click(struct litest_device *d, unsigned int button, bool is_press)
 		litest_event(d, ev->type, ev->code, ev->value);
 }
 
+void
+litest_button_scroll(struct litest_device *dev,
+		     unsigned int button,
+		     double dx, double dy)
+{
+	struct libinput *li = dev->libinput;
+
+	litest_button_click(dev, button, 1);
+
+	libinput_dispatch(li);
+	litest_timeout_buttonscroll();
+	libinput_dispatch(li);
+
+	litest_event(dev, EV_REL, REL_X, dx);
+	litest_event(dev, EV_REL, REL_Y, dy);
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+
+	litest_button_click(dev, button, 0);
+
+	libinput_dispatch(li);
+}
+
 void
 litest_keyboard_key(struct litest_device *d, unsigned int key, bool is_press)
 {
@@ -958,7 +1014,11 @@ litest_create_uinput_device_from_description(const char *name,
 	rc = libevdev_uinput_create_from_device(dev,
 					        LIBEVDEV_UINPUT_OPEN_MANAGED,
 						&uinput);
-	ck_assert_int_eq(rc, 0);
+	/* workaround for a bug in libevdev pre-1.3
+	   http://cgit.freedesktop.org/libevdev/commit/?id=debe9b030c8069cdf78307888ef3b65830b25122 */
+	if (rc == -EBADF)
+		rc = -EACCES;
+	ck_assert_msg(rc == 0, "Failed to create uinput device: %s", strerror(-rc));
 
 	libevdev_free(dev);
 
@@ -1042,3 +1102,99 @@ litest_create_uinput_device(const char *name, struct input_id *id, ...)
 
 	return uinput;
 }
+
+void
+litest_assert_button_event(struct libinput *li, unsigned int button,
+			   enum libinput_button_state state)
+{
+	struct libinput_event *event;
+	struct libinput_event_pointer *ptrev;
+
+	litest_wait_for_event(li);
+	event = libinput_get_event(li);
+
+	ck_assert(event != NULL);
+	ck_assert_int_eq(libinput_event_get_type(event),
+			 LIBINPUT_EVENT_POINTER_BUTTON);
+	ptrev = libinput_event_get_pointer_event(event);
+	ck_assert_int_eq(libinput_event_pointer_get_button(ptrev),
+			 button);
+	ck_assert_int_eq(libinput_event_pointer_get_button_state(ptrev),
+			 state);
+	libinput_event_destroy(event);
+}
+
+void
+litest_assert_scroll(struct libinput *li,
+		     enum libinput_pointer_axis axis,
+		     int minimum_movement)
+{
+	struct libinput_event *event, *next_event;
+	struct libinput_event_pointer *ptrev;
+
+	event = libinput_get_event(li);
+	next_event = libinput_get_event(li);
+	ck_assert(next_event != NULL); /* At least 1 scroll + stop scroll */
+
+	while (event) {
+		ck_assert_int_eq(libinput_event_get_type(event),
+				 LIBINPUT_EVENT_POINTER_AXIS);
+		ptrev = libinput_event_get_pointer_event(event);
+		ck_assert(ptrev != NULL);
+		ck_assert_int_eq(libinput_event_pointer_get_axis(ptrev), axis);
+
+		if (next_event) {
+			/* Normal scroll event, check dir */
+			if (minimum_movement > 0) {
+				ck_assert_int_ge(
+					libinput_event_pointer_get_axis_value(ptrev),
+					minimum_movement);
+			} else {
+				ck_assert_int_le(
+					libinput_event_pointer_get_axis_value(ptrev),
+					minimum_movement);
+			}
+		} else {
+			/* Last scroll event, must be 0 */
+			ck_assert_int_eq(
+				libinput_event_pointer_get_axis_value(ptrev),
+				0);
+		}
+		libinput_event_destroy(event);
+		event = next_event;
+		next_event = libinput_get_event(li);
+	}
+}
+
+void
+litest_timeout_tap(void)
+{
+	msleep(200);
+}
+
+void
+litest_timeout_softbuttons(void)
+{
+	msleep(300);
+}
+
+void
+litest_timeout_buttonscroll(void)
+{
+	msleep(300);
+}
+
+void
+litest_push_event_frame(struct litest_device *dev)
+{
+	assert(!dev->skip_ev_syn);
+	dev->skip_ev_syn = true;
+}
+
+void
+litest_pop_event_frame(struct litest_device *dev)
+{
+	assert(dev->skip_ev_syn);
+	dev->skip_ev_syn = false;
+	litest_event(dev, EV_SYN, SYN_REPORT, 0);
+}
