@@ -391,25 +391,6 @@ START_TEST(context_ref_counting)
 }
 END_TEST
 
-START_TEST(device_ids)
-{
-	struct litest_device *dev = litest_current_device();
-	const char *name;
-	unsigned int pid, vid;
-
-	name = libevdev_get_name(dev->evdev);
-	pid = libevdev_get_id_product(dev->evdev);
-	vid = libevdev_get_id_vendor(dev->evdev);
-
-	ck_assert_str_eq(name,
-			 libinput_device_get_name(dev->libinput_device));
-	ck_assert_int_eq(pid,
-			 libinput_device_get_id_product(dev->libinput_device));
-	ck_assert_int_eq(vid,
-			 libinput_device_get_id_vendor(dev->libinput_device));
-}
-END_TEST
-
 START_TEST(config_status_string)
 {
 	const char *strs[3];
@@ -508,6 +489,87 @@ START_TEST(matrix_helpers)
 }
 END_TEST
 
+START_TEST(ratelimit_helpers)
+{
+	struct ratelimit rl;
+	unsigned int i, j;
+
+	/* 10 attempts every 100ms */
+	ratelimit_init(&rl, 100, 10);
+
+	for (j = 0; j < 3; ++j) {
+		/* a burst of 9 attempts must succeed */
+		for (i = 0; i < 9; ++i) {
+			ck_assert_int_eq(ratelimit_test(&rl),
+					 RATELIMIT_PASS);
+		}
+
+		/* the 10th attempt reaches the threshold */
+		ck_assert_int_eq(ratelimit_test(&rl), RATELIMIT_THRESHOLD);
+
+		/* ..then further attempts must fail.. */
+		ck_assert_int_eq(ratelimit_test(&rl), RATELIMIT_EXCEEDED);
+
+		/* ..regardless of how often we try. */
+		for (i = 0; i < 100; ++i) {
+			ck_assert_int_eq(ratelimit_test(&rl),
+					 RATELIMIT_EXCEEDED);
+		}
+
+		/* ..even after waiting 20ms */
+		msleep(20);
+		for (i = 0; i < 100; ++i) {
+			ck_assert_int_eq(ratelimit_test(&rl),
+					 RATELIMIT_EXCEEDED);
+		}
+
+		/* but after 100ms the counter is reset */
+		msleep(90); /* +10ms to account for time drifts */
+	}
+}
+END_TEST
+
+struct parser_test {
+	char *tag;
+	int expected_dpi;
+};
+
+START_TEST(dpi_parser)
+{
+	struct parser_test tests[] = {
+		{ "450 *1800 3200", 1800 },
+		{ "*450 1800 3200", 450 },
+		{ "450 1800 *3200", 3200 },
+		{ "450 1800 3200", 3200 },
+		{ "450 1800 failboat", 0 },
+		{ "450 1800 *failboat", 0 },
+		{ "0 450 1800 *3200", 0 },
+		{ "450@37 1800@12 *3200@6", 3200 },
+		{ "450@125 1800@125   *3200@125  ", 3200 },
+		{ "450@125 *1800@125  3200@125", 1800 },
+		{ "*this @string fails", 0 },
+		{ "12@34 *45@", 0 },
+		{ "12@a *45@", 0 },
+		{ "12@a *45@25", 0 },
+		{ "                                      * 12, 450, 800", 0 },
+		{ "                                      *12, 450, 800", 12 },
+		{ "*12, *450, 800", 12 },
+		{ "*-23412, 450, 800", 0 },
+		{ "112@125, 450@125, 800@125, 900@-125", 0 },
+		{ "", 0 },
+		{ "   ", 0 },
+		{ "* ", 0 },
+		{ NULL }
+	};
+	int i, dpi;
+
+	for (i = 0; tests[i].tag != NULL; i++) {
+		dpi = parse_mouse_dpi_property(tests[i].tag);
+		ck_assert_int_eq(dpi, tests[i].expected_dpi);
+	}
+}
+END_TEST
+
 int main (int argc, char **argv) {
 	litest_add_no_device("events:conversion", event_conversion_device_notify);
 	litest_add_no_device("events:conversion", event_conversion_pointer);
@@ -515,9 +577,11 @@ int main (int argc, char **argv) {
 	litest_add_no_device("events:conversion", event_conversion_key);
 	litest_add_no_device("events:conversion", event_conversion_touch);
 	litest_add_no_device("context:refcount", context_ref_counting);
-	litest_add("device:id", device_ids, LITEST_ANY, LITEST_ANY);
 	litest_add_no_device("config:status string", config_status_string);
 
 	litest_add_no_device("misc:matrix", matrix_helpers);
+	litest_add_no_device("misc:ratelimit", ratelimit_helpers);
+	litest_add_no_device("misc:dpi parser", dpi_parser);
+
 	return litest_run(argc, argv);
 }
