@@ -67,6 +67,8 @@ struct window {
 
 	/* l/m/r mouse buttons */
 	int l, m, r;
+
+	struct libinput_device *devices[50];
 };
 
 static int
@@ -211,18 +213,69 @@ window_init(struct window *w)
 	gtk_widget_show_all(w->win);
 }
 
+static void
+window_cleanup(struct window *w)
+{
+	struct libinput_device **dev;
+	ARRAY_FOR_EACH(w->devices, dev) {
+		if (*dev)
+			libinput_device_unref(*dev);
+	}
+}
+
+static void
+change_ptraccel(struct window *w, double amount)
+{
+	struct libinput_device **dev;
+
+	ARRAY_FOR_EACH(w->devices, dev) {
+		double speed;
+		enum libinput_config_status status;
+
+		if (*dev == NULL)
+			continue;
+
+		if (!libinput_device_config_accel_is_available(*dev))
+			continue;
+
+		speed = libinput_device_config_accel_get_speed(*dev);
+		speed = clip(speed + amount, -1, 1);
+
+		status = libinput_device_config_accel_set_speed(*dev, speed);
+
+		if (status != LIBINPUT_CONFIG_STATUS_SUCCESS) {
+			msg("%s: failed to change accel to %.2f (%s)\n",
+			    libinput_device_get_name(*dev),
+			    speed,
+			    libinput_config_status_to_str(status));
+		} else {
+			printf("%s: speed is %.2f\n",
+			       libinput_device_get_name(*dev),
+			       speed);
+		}
+
+	}
+}
+
+
 static void
 handle_event_device_notify(struct libinput_event *ev)
 {
 	struct libinput_device *dev = libinput_event_get_device(ev);
+	struct libinput *li;
+	struct window *w;
 	const char *type;
+	int i;
 
 	if (libinput_event_get_type(ev) == LIBINPUT_EVENT_DEVICE_ADDED)
 		type = "added";
 	else
 		type = "removed";
 
-	msg("%s %s\n", libinput_device_get_sysname(dev), type);
+	msg("%s %-30s %s\n",
+	    libinput_device_get_sysname(dev),
+	    libinput_device_get_name(dev),
+	    type);
 
 	if (libinput_device_config_tap_get_finger_count(dev) > 0) {
 		enum libinput_config_status status;
@@ -232,6 +285,26 @@ handle_event_device_notify(struct libinput_event *ev)
 			error("%s: Failed to enable tapping\n",
 			      libinput_device_get_sysname(dev));
 	}
+
+	li = libinput_event_get_context(ev);
+	w = libinput_get_user_data(li);
+
+	if (libinput_event_get_type(ev) == LIBINPUT_EVENT_DEVICE_ADDED) {
+		for (i = 0; i < ARRAY_LENGTH(w->devices); i++) {
+			if (w->devices[i] == NULL) {
+				w->devices[i] = libinput_device_ref(dev);
+				break;
+			}
+		}
+	} else  {
+		for (i = 0; i < ARRAY_LENGTH(w->devices); i++) {
+			if (w->devices[i] == dev) {
+				libinput_device_unref(w->devices[i]);
+				w->devices[i] = NULL;
+				break;
+			}
+		}
+	}
 }
 
 static void
@@ -309,9 +382,24 @@ static int
 handle_event_keyboard(struct libinput_event *ev, struct window *w)
 {
 	struct libinput_event_keyboard *k = libinput_event_get_keyboard_event(ev);
+	unsigned int key = libinput_event_keyboard_get_key(k);
 
-	if (libinput_event_keyboard_get_key(k) == KEY_ESC)
+	if (libinput_event_keyboard_get_key_state(k) ==
+	    LIBINPUT_KEY_STATE_RELEASED)
+		return 0;
+
+	switch(key) {
+	case KEY_ESC:
 		return 1;
+	case KEY_UP:
+		change_ptraccel(w, 0.1);
+		break;
+	case KEY_DOWN:
+		change_ptraccel(w, -0.1);
+		break;
+	default:
+		break;
+	}
 
 	return 0;
 }
@@ -474,6 +562,7 @@ main(int argc, char *argv[])
 
 	gtk_main();
 
+	window_cleanup(&w);
 	libinput_unref(li);
 	udev_unref(udev);
 
